@@ -1,4 +1,3 @@
-# -*- coding: utf-8 -*-
 """
     flask_security
     ~~~~~~~~~~~~~~
@@ -47,9 +46,11 @@ from .forms import (
     TwoFactorRescueForm,
     TwoFactorSetupForm,
     TwoFactorVerifyCodeForm,
-    TwoFactorVerifyPasswordForm,
     VerifyForm,
+    unique_identity_attribute,
 )
+from .mail_util import MailUtil
+from .password_util import PasswordUtil
 from .phone_util import PhoneUtil
 from .signals import (
     confirm_instructions_sent,
@@ -82,6 +83,7 @@ from .utils import (
     SmsSenderFactory,
     check_and_get_token_status,
     get_hmac,
+    get_request_attr,
     get_token_status,
     get_url,
     hash_password,
@@ -101,4 +103,4 @@ from .utils import (
     verify_and_update_password,
 )
 
-__version__ = "3.4.5"
+__version__ = "4.0.1"
