@@ -1,4 +1,3 @@
-# -*- coding: utf-8 -*-
 """
     conftest
     ~~~~~~~~
@@ -13,20 +12,14 @@ import os
 import tempfile
 import time
 from datetime import datetime
-
-try:
-    from urlparse import urlsplit
-except ImportError:  # pragma: no cover
-    from urllib.parse import urlsplit
+from urllib.parse import urlsplit
 
 import pytest
 from flask import Flask, Response, render_template
 from flask import jsonify
 from flask import request as flask_request
 from flask.json import JSONEncoder
-from flask_babelex import Babel
 from flask_mail import Mail
-from utils import populate_data
 
 from flask_security import (
     MongoEngineUserDatastore,
@@ -40,12 +33,26 @@ from flask_security import (
     auth_required,
     auth_token_required,
     http_auth_required,
+    get_request_attr,
     login_required,
     roles_accepted,
     roles_required,
     permissions_accepted,
     permissions_required,
+    uia_email_mapper,
 )
+from flask_security.utils import localize_callback
+
+from tests.test_utils import populate_data
+
+NO_BABEL = False
+try:
+    from flask_babel import Babel
+except ImportError:
+    try:
+        from flask_babelex import Babel
+    except ImportError:
+        NO_BABEL = True
 
 
 @pytest.fixture()
@@ -57,6 +64,8 @@ def app(request):
     app.config["TESTING"] = True
     app.config["LOGIN_DISABLED"] = False
     app.config["WTF_CSRF_ENABLED"] = False
+    # Our test emails/domain isn't necessarily valid
+    app.config["SECURITY_EMAIL_VALIDATOR_ARGS"] = {"check_deliverability": False}
     app.config["SECURITY_TWO_FACTOR_SECRET"] = {
         "1": "TjQ9Qa31VOrfEzuPy4VHQWPCTmRzCnFzMKLxXYiZu9B"
     }
@@ -94,7 +103,7 @@ def app(request):
             app.config["SECURITY_" + key.upper()] = value
 
     mail = Mail(app)
-    if babel is None or babel.args[0]:
+    if not NO_BABEL and (babel is None or babel.args[0]):
         Babel(app)
     app.json_encoder = JSONEncoder
     app.mail = mail
@@ -126,16 +135,19 @@ def app(request):
     @http_auth_required
     @permissions_required("admin")
     def http_admin_required():
+        assert get_request_attr("fs_authn_via") == "basic"
         return "HTTP Authentication"
 
     @app.route("/http_custom_realm")
     @http_auth_required("My Realm")
     def http_custom_realm():
+        assert get_request_attr("fs_authn_via") == "basic"
         return render_template("index.html", content="HTTP Authentication")
 
     @app.route("/token", methods=["GET", "POST"])
     @auth_token_required
     def token():
+        assert get_request_attr("fs_authn_via") == "token"
         return render_template("index.html", content="Token Authentication")
 
     @app.route("/multi_auth")
@@ -158,6 +170,7 @@ def app(request):
     @app.route("/admin")
     @roles_required("admin")
     def admin():
+        assert get_request_attr("fs_authn_via") == "session"
         return render_template("index.html", content="Admin Page")
 
     @app.route("/admin_and_editor")
@@ -187,10 +200,6 @@ def app(request):
     def admin_perm_required():
         return render_template("index.html", content="Admin Page required")
 
-    @app.route("/unauthorized")
-    def unauthorized():
-        return render_template("unauthorized.html")
-
     @app.route("/page1")
     def page_1():
         return "Page 1"
@@ -220,7 +229,16 @@ def mongoengine_datastore(request, app, tmpdir, realdburl):
 
 
 def mongoengine_setup(request, app, tmpdir, realdburl):
+    pytest.importorskip("flask_mongoengine")
     from flask_mongoengine import MongoEngine
+    from mongoengine.fields import (
+        BooleanField,
+        DateTimeField,
+        IntField,
+        ListField,
+        ReferenceField,
+        StringField,
+    )
 
     db_name = "flask_security_test_%s" % str(time.time()).replace(".", "_")
     app.config["MONGODB_SETTINGS"] = {
@@ -233,29 +251,30 @@ def mongoengine_setup(request, app, tmpdir, realdburl):
     db = MongoEngine(app)
 
     class Role(db.Document, RoleMixin):
-        name = db.StringField(required=True, unique=True, max_length=80)
-        description = db.StringField(max_length=255)
-        permissions = db.StringField(max_length=255)
+        name = StringField(required=True, unique=True, max_length=80)
+        description = StringField(max_length=255)
+        permissions = StringField(max_length=255)
         meta = {"db_alias": db_name}
 
     class User(db.Document, UserMixin):
-        email = db.StringField(unique=True, max_length=255)
-        username = db.StringField(unique=True, max_length=255)
-        password = db.StringField(required=False, max_length=255)
-        security_number = db.IntField(unique=True)
-        last_login_at = db.DateTimeField()
-        current_login_at = db.DateTimeField()
-        tf_primary_method = db.StringField(max_length=255)
-        tf_totp_secret = db.StringField(max_length=255)
-        tf_phone_number = db.StringField(max_length=255)
-        us_totp_secrets = db.StringField()
-        us_phone_number = db.StringField(max_length=255)
-        last_login_ip = db.StringField(max_length=100)
-        current_login_ip = db.StringField(max_length=100)
-        login_count = db.IntField()
-        active = db.BooleanField(default=True)
-        confirmed_at = db.DateTimeField()
-        roles = db.ListField(db.ReferenceField(Role), default=[])
+        email = StringField(unique=True, max_length=255)
+        fs_uniquifier = StringField(unique=True, max_length=64, required=True)
+        username = StringField(unique=True, required=False, sparse=True, max_length=255)
+        password = StringField(required=False, max_length=255)
+        security_number = IntField(unique=True, required=False, sparse=True)
+        last_login_at = DateTimeField()
+        current_login_at = DateTimeField()
+        tf_primary_method = StringField(max_length=255)
+        tf_totp_secret = StringField(max_length=255)
+        tf_phone_number = StringField(max_length=255)
+        us_totp_secrets = StringField()
+        us_phone_number = StringField(max_length=255)
+        last_login_ip = StringField(max_length=100)
+        current_login_ip = StringField(max_length=100)
+        login_count = IntField()
+        active = BooleanField(default=True)
+        confirmed_at = DateTimeField()
+        roles = ListField(ReferenceField(Role), default=[])
         meta = {"db_alias": db_name}
 
     def tear_down():
@@ -275,6 +294,7 @@ def sqlalchemy_datastore(request, app, tmpdir, realdburl):
 
 
 def sqlalchemy_setup(request, app, tmpdir, realdburl):
+    pytest.importorskip("flask_sqlalchemy")
     from flask_sqlalchemy import SQLAlchemy
     from flask_security.models import fsqla_v2 as fsqla
 
@@ -298,7 +318,7 @@ def sqlalchemy_setup(request, app, tmpdir, realdburl):
 
         def get_security_payload(self):
             # Make sure we still properly hook up to flask JSONEncoder
-            return {"id": str(self.id), "last_update": self.update_datetime}
+            return {"email": str(self.email), "last_update": self.update_datetime}
 
     with app.app_context():
         db.create_all()
@@ -319,10 +339,21 @@ def sqlalchemy_session_datastore(request, app, tmpdir, realdburl):
 
 
 def sqlalchemy_session_setup(request, app, tmpdir, realdburl):
+    pytest.importorskip("sqlalchemy")
     from sqlalchemy import create_engine
     from sqlalchemy.orm import scoped_session, sessionmaker, relationship, backref
     from sqlalchemy.ext.declarative import declarative_base
-    from sqlalchemy import Boolean, DateTime, Column, Integer, String, Text, ForeignKey
+    from sqlalchemy.sql import func
+    from sqlalchemy import (
+        Boolean,
+        DateTime,
+        Column,
+        Integer,
+        String,
+        Text,
+        ForeignKey,
+        UnicodeText,
+    )
 
     f, path = tempfile.mkstemp(
         prefix="flask-security-test-db", suffix=".db", dir=str(tmpdir)
@@ -330,7 +361,7 @@ def sqlalchemy_session_setup(request, app, tmpdir, realdburl):
 
     app.config["SQLALCHEMY_DATABASE_URI"] = "sqlite:///" + path
 
-    engine = create_engine(app.config["SQLALCHEMY_DATABASE_URI"], convert_unicode=True)
+    engine = create_engine(app.config["SQLALCHEMY_DATABASE_URI"])
     db_session = scoped_session(
         sessionmaker(autocommit=False, autoflush=False, bind=engine)
     )
@@ -340,21 +371,28 @@ def sqlalchemy_session_setup(request, app, tmpdir, realdburl):
     class RolesUsers(Base):
         __tablename__ = "roles_users"
         id = Column(Integer(), primary_key=True)
-        user_id = Column("user_id", Integer(), ForeignKey("user.id"))
-        role_id = Column("role_id", Integer(), ForeignKey("role.id"))
+        user_id = Column("user_id", Integer(), ForeignKey("user.myuserid"))
+        role_id = Column("role_id", Integer(), ForeignKey("role.myroleid"))
 
     class Role(Base, RoleMixin):
         __tablename__ = "role"
-        id = Column(Integer(), primary_key=True)
+        myroleid = Column(Integer(), primary_key=True)
         name = Column(String(80), unique=True)
         description = Column(String(255))
-        permissions = Column(String(255))
+        permissions = Column(UnicodeText, nullable=True)
+        update_datetime = Column(
+            DateTime,
+            nullable=False,
+            server_default=func.now(),
+            onupdate=datetime.utcnow,
+        )
 
     class User(Base, UserMixin):
         __tablename__ = "user"
-        id = Column(Integer, primary_key=True)
+        myuserid = Column(Integer, primary_key=True)
+        fs_uniquifier = Column(String(64), unique=True, nullable=False)
         email = Column(String(255), unique=True)
-        username = Column(String(255))
+        username = Column(String(255), unique=True, nullable=True)
         password = Column(String(255))
         security_number = Column(Integer, unique=True)
         last_login_at = Column(DateTime())
@@ -372,6 +410,16 @@ def sqlalchemy_session_setup(request, app, tmpdir, realdburl):
         roles = relationship(
             "Role", secondary="roles_users", backref=backref("users", lazy="dynamic")
         )
+        update_datetime = Column(
+            DateTime,
+            nullable=False,
+            server_default=func.now(),
+            onupdate=datetime.utcnow,
+        )
+
+        def get_security_payload(self):
+            # Make sure we still properly hook up to flask JSONEncoder
+            return {"email": str(self.email), "last_update": self.update_datetime}
 
     with app.app_context():
         Base.metadata.create_all(bind=engine)
@@ -392,6 +440,7 @@ def peewee_datastore(request, app, tmpdir, realdburl):
 
 
 def peewee_setup(request, app, tmpdir, realdburl):
+    pytest.importorskip("peewee")
     from peewee import (
         TextField,
         DateTimeField,
@@ -413,8 +462,9 @@ def peewee_setup(request, app, tmpdir, realdburl):
             "name": pieces.path[1:],
             "engine": engine_mapper[pieces.scheme.split("+")[0]],
             "user": pieces.username,
-            "passwd": pieces.password,
+            "password": pieces.password,
             "host": pieces.hostname,
+            "port": pieces.port,
         }
     else:
         f, path = tempfile.mkstemp(
@@ -432,9 +482,9 @@ def peewee_setup(request, app, tmpdir, realdburl):
         permissions = TextField(null=True)
 
     class User(UserMixin, db.Model):
-        email = TextField()
+        email = TextField(unique=True, null=False)
         fs_uniquifier = TextField(unique=True, null=False)
-        username = TextField()
+        username = TextField(unique=True, null=True)
         security_number = IntegerField(null=True)
         password = TextField(null=True)
         last_login_at = DateTimeField(null=True)
@@ -451,7 +501,7 @@ def peewee_setup(request, app, tmpdir, realdburl):
         confirmed_at = DateTimeField(null=True)
 
     class UserRoles(db.Model):
-        """ Peewee does not have built-in many-to-many support, so we have to
+        """Peewee does not have built-in many-to-many support, so we have to
         create this mapping class to link users to roles."""
 
         user = ForeignKeyField(User, backref="roles")
@@ -488,6 +538,7 @@ def pony_datastore(request, app, tmpdir, realdburl):
 
 def pony_setup(request, app, tmpdir, realdburl):
 
+    pytest.importorskip("pony")
     from pony.orm import Database, Optional, Required, Set
     from pony.orm.core import SetInstance
 
@@ -501,6 +552,7 @@ def pony_setup(request, app, tmpdir, realdburl):
 
     class User(db.Entity):
         email = Required(str)
+        fs_uniquifier = Required(str, nullable=False)
         username = Optional(str)
         security_number = Optional(int)
         password = Optional(str, nullable=True)
@@ -529,6 +581,7 @@ def pony_setup(request, app, tmpdir, realdburl):
             user=pieces.username,
             password=pieces.password,
             host=pieces.hostname,
+            port=pieces.port,
             database=pieces.path[1:],
         )
     else:
@@ -628,7 +681,7 @@ def clients(request, app, tmpdir, realdburl):
     return app.test_client()
 
 
-@pytest.yield_fixture()
+@pytest.fixture()
 def in_app_context(request, sqlalchemy_app):
     app = sqlalchemy_app()
     with app.app_context():
@@ -644,6 +697,14 @@ def get_message(app):
     return fn
 
 
+@pytest.fixture()
+def get_message_local(app):
+    def fn(key, **kwargs):
+        return localize_callback(app.config["SECURITY_MSG_" + key][0], **kwargs)
+
+    return fn
+
+
 @pytest.fixture(
     params=["sqlalchemy", "sqlalchemy-session", "mongoengine", "peewee", "pony"]
 )
@@ -662,17 +723,18 @@ def datastore(request, app, tmpdir, realdburl):
 
 
 @pytest.fixture()
-def script_info(app, datastore):
-    try:
-        from flask.cli import ScriptInfo
-    except ImportError:
-        from flask_cli import ScriptInfo
+# def script_info(app, datastore): # Fix me when pony works
+def script_info(app, sqlalchemy_datastore):
+    from flask.cli import ScriptInfo
 
-    def create_app(info):
-        app.config.update(
-            **{"SECURITY_USER_IDENTITY_ATTRIBUTES": ("email", "username")}
-        )
-        app.security = Security(app, datastore=datastore)
+    def create_app():
+        uia = [
+            {"email": {"mapper": uia_email_mapper}},
+            {"username": {"mapper": lambda x: x}},
+        ]
+
+        app.config.update(**{"SECURITY_USER_IDENTITY_ATTRIBUTES": uia})
+        app.security = Security(app, datastore=sqlalchemy_datastore)
         return app
 
     return ScriptInfo(create_app=create_app)
