@@ -1,4 +1,3 @@
-# -*- coding: utf-8 -*-
 """
     test_common
     ~~~~~~~~~~~
@@ -11,24 +10,25 @@
 
 import base64
 import json
+import re
+from http.cookiejar import Cookie
 import pytest
 
 from flask import Blueprint
 
-from utils import (
+from flask_security import uia_email_mapper
+from flask_security.utils import hash_data
+
+from tests.test_utils import (
     authenticate,
     json_authenticate,
     get_num_queries,
+    hash_password,
     logout,
     populate_data,
     verify_token,
 )
 
-try:
-    from cookielib import Cookie
-except ImportError:
-    from http.cookiejar import Cookie
-
 
 def test_login_view(client):
     response = client.get("/login")
@@ -74,6 +74,38 @@ def test_authenticate_with_invalid_malformed_next(client, get_message):
     assert get_message("INVALID_REDIRECT") in response.data
 
 
+def test_authenticate_with_subdomain_next(app, client, get_message):
+    app.config["SERVER_NAME"] = "lp.com"
+    app.config["SECURITY_REDIRECT_ALLOW_SUBDOMAINS"] = True
+    data = dict(email="matt@lp.com", password="password")
+    response = client.post("/login?next=http://sub.lp.com", data=data)
+    assert response.status_code == 302
+
+
+def test_authenticate_with_root_domain_next(app, client, get_message):
+    app.config["SERVER_NAME"] = "lp.com"
+    app.config["SECURITY_SUBDOMAIN"] = "auth"
+    app.config["SECURITY_REDIRECT_ALLOW_SUBDOMAINS"] = True
+    data = dict(email="matt@lp.com", password="password")
+    response = client.post("/login?next=http://lp.com", data=data)
+    assert response.status_code == 302
+
+
+def test_authenticate_with_invalid_subdomain_next(app, client, get_message):
+    app.config["SERVER_NAME"] = "lp.com"
+    app.config["SECURITY_REDIRECT_ALLOW_SUBDOMAINS"] = True
+    data = dict(email="matt@lp.com", password="password")
+    response = client.post("/login?next=http://sub.lp.net", data=data)
+    assert get_message("INVALID_REDIRECT") in response.data
+
+
+def test_authenticate_with_subdomain_next_default_config(app, client, get_message):
+    app.config["SERVER_NAME"] = "lp.com"
+    data = dict(email="matt@lp.com", password="password")
+    response = client.post("/login?next=http://sub.lp.com", data=data)
+    assert get_message("INVALID_REDIRECT") in response.data
+
+
 def test_authenticate_case_insensitive_email(app, client):
     response = authenticate(client, "MATT@lp.com", follow_redirects=True)
     assert b"Welcome matt@lp.com" in response.data
@@ -118,6 +150,7 @@ def test_post_already_authenticated(client):
 def test_login_form(client):
     response = client.post("/login", data={"email": "matt@lp.com"})
     assert b"matt@lp.com" in response.data
+    assert re.search(b'<input[^>]*type="email"[^>]*>', response.data)
 
 
 def test_unprovided_username(client, get_message):
@@ -146,7 +179,7 @@ def test_inactive_user(client, get_message):
 
 
 def test_inactive_forbids(app, client, get_message):
-    """ Make sure that existing session doesn't work after
+    """Make sure that existing session doesn't work after
     user marked inactive
     """
     response = authenticate(client, follow_redirects=True)
@@ -169,7 +202,7 @@ def test_inactive_forbids(app, client, get_message):
 
 @pytest.mark.settings(unauthorized_view=None)
 def test_inactive_forbids_token(app, client_nc, get_message):
-    """ Make sure that existing token doesn't work after
+    """Make sure that existing token doesn't work after
     user marked inactive
     """
     response = json_authenticate(client_nc)
@@ -190,6 +223,35 @@ def test_inactive_forbids_token(app, client_nc, get_message):
     assert response.status_code == 401
 
 
+def test_inactive_forbids_basic(app, client, get_message):
+    """Make sure that basic auth doesn't work if user deactivated"""
+
+    # Should properly work.
+    response = client.get(
+        "/multi_auth",
+        headers={
+            "Authorization": "Basic %s"
+            % base64.b64encode(b"joe@lp.com:password").decode("utf-8")
+        },
+    )
+    assert b"Session, Token, Basic" in response.data
+
+    # deactivate joe
+    with app.test_request_context("/"):
+        user = app.security.datastore.find_user(email="joe@lp.com")
+        app.security.datastore.deactivate_user(user)
+        app.security.datastore.commit()
+
+    response = client.get(
+        "/multi_auth",
+        headers={
+            "Authorization": "Basic %s"
+            % base64.b64encode(b"joe@lp.com:password").decode("utf-8")
+        },
+    )
+    assert b"You are not authenticated" in response.data
+
+
 def test_unset_password(client, get_message):
     response = authenticate(client, "jess@lp.com", "password")
     assert get_message("PASSWORD_NOT_SET") in response.data
@@ -385,7 +447,12 @@ def test_http_auth(client):
     assert b"HTTP Authentication" in response.data
 
 
-@pytest.mark.settings(USER_IDENTITY_ATTRIBUTES=("email", "username"))
+@pytest.mark.settings(
+    USER_IDENTITY_ATTRIBUTES=[
+        {"email": {"mapper": uia_email_mapper}},
+        {"username": {"mapper": lambda x: x}},
+    ]
+)
 def test_http_auth_username(client):
     response = client.get(
         "/http",
@@ -438,7 +505,6 @@ def test_http_auth_no_authentication_json(client, get_message):
         "UNAUTHENTICATED"
     )
     assert response.headers["Content-Type"] == "application/json"
-    assert "WWW-Authenticate" not in response.headers
 
 
 @pytest.mark.settings(backwards_compat_unauthn=True)
@@ -457,9 +523,7 @@ def test_invalid_http_auth_invalid_username(client):
 
 @pytest.mark.settings(backwards_compat_unauthn=False)
 def test_invalid_http_auth_invalid_username_json(client, get_message):
-    # While Basic auth is allowed with JSON - we never expect a WWW-Authenticate
-    # header - since that is captured by most browsers and they pop up a
-    # login form.
+    # Even with JSON - Basic Auth required a WWW-Authenticate header response.
     response = client.get(
         "/http",
         headers={
@@ -473,7 +537,7 @@ def test_invalid_http_auth_invalid_username_json(client, get_message):
         "UNAUTHENTICATED"
     )
     assert response.headers["Content-Type"] == "application/json"
-    assert "WWW-Authenticate" not in response.headers
+    assert "WWW-Authenticate" in response.headers
 
 
 @pytest.mark.settings(backwards_compat_unauthn=True)
@@ -635,12 +699,12 @@ def test_login_info(client):
     json_authenticate(client)
     response = client.get("/login", headers={"Content-Type": "application/json"})
     assert response.status_code == 200
-    assert response.json["response"]["user"]["id"] == "1"
+    assert response.json["response"]["user"]["email"] == "matt@lp.com"
     assert "last_update" in response.json["response"]["user"]
 
     response = client.get("/login", headers={"Accept": "application/json"})
     assert response.status_code == 200
-    assert response.json["response"]["user"]["id"] == "1"
+    assert response.json["response"]["user"]["email"] == "matt@lp.com"
     assert "last_update" in response.json["response"]["user"]
 
 
@@ -666,27 +730,6 @@ def test_anon_required_json(client, get_message):
     )
 
 
-@pytest.mark.settings(security_hashing_schemes=["sha256_crypt"])
-@pytest.mark.skip
-def test_auth_token_speed(app, client_nc):
-    # To run with old algorithm you have to comment out fs_uniquifier check in UserMixin
-    import timeit
-
-    response = json_authenticate(client_nc)
-    token = response.json["response"]["user"]["authentication_token"]
-
-    def time_get():
-        rp = client_nc.get(
-            "/login",
-            data={},
-            headers={"Content-Type": "application/json", "Authentication-Token": token},
-        )
-        assert rp.status_code == 200
-
-    t = timeit.timeit(time_get, number=50)
-    print("Time for 50 iterations: ", t)
-
-
 def test_change_uniquifier(app, client_nc):
     # make sure that existing token no longer works once we change the uniquifier
 
@@ -708,6 +751,128 @@ def test_change_uniquifier(app, client_nc):
     verify_token(client_nc, token)
 
 
+def test_verifying_token_from_version_3x(app, client_nc):
+    """
+    Check token generated with flask security 3.x, which has different form
+    than token from version 4.0.0, can be verified
+    """
+
+    def get_auth_token_version_3x(app, user):
+        """
+        Copy of algorithm that generated user token in version 3.x
+        """
+        data = [str(user.id), hash_data(user.password)]
+        if hasattr(user, "fs_uniquifier"):
+            data.append(user.fs_uniquifier)
+        return app.security.remember_token_serializer.dumps(data)
+
+    with app.test_request_context("/"):
+        user = app.security.datastore.find_user(email="matt@lp.com")
+
+        token = get_auth_token_version_3x(app, user)
+
+        data = app.security.remember_token_serializer.loads(
+            token, max_age=app.security.token_max_age
+        )
+
+        assert user.verify_auth_token(data) is True
+
+
+def test_change_token_uniquifier(app):
+    # make sure that existing token no longer works once we change the token uniquifier
+    from sqlalchemy import Column, String
+    from flask_sqlalchemy import SQLAlchemy
+    from flask_security.models import fsqla_v2 as fsqla
+    from flask_security import Security, SQLAlchemyUserDatastore
+
+    app.config["SQLALCHEMY_DATABASE_URI"] = "sqlite:///:memory:"
+    db = SQLAlchemy(app)
+
+    fsqla.FsModels.set_db_info(db)
+
+    class Role(db.Model, fsqla.FsRoleMixin):
+        pass
+
+    class User(db.Model, fsqla.FsUserMixin):
+        fs_token_uniquifier = Column(String(64), unique=True, nullable=False)
+
+    with app.app_context():
+        db.create_all()
+
+    ds = SQLAlchemyUserDatastore(db, User, Role)
+    app.security = Security(app, datastore=ds)
+
+    with app.app_context():
+        ds.create_user(
+            email="matt@lp.com",
+            password=hash_password("password"),
+        )
+        ds.commit()
+
+        client_nc = app.test_client(use_cookies=False)
+
+        response = json_authenticate(client_nc)
+        token = response.json["response"]["user"]["authentication_token"]
+        verify_token(client_nc, token)
+
+        # now change uniquifier
+        with app.test_request_context("/"):
+            user = app.security.datastore.find_user(email="matt@lp.com")
+            app.security.datastore.reset_user_access(user)
+            app.security.datastore.commit()
+
+        verify_token(client_nc, token, status=401)
+
+        # get new token and verify it works
+        response = json_authenticate(client_nc)
+        token = response.json["response"]["user"]["authentication_token"]
+        verify_token(client_nc, token)
+
+
+def test_null_token_uniquifier(app):
+    # If existing record has a null fs_token_uniquifier, should be set on first use.
+    from sqlalchemy import Column, String
+    from flask_sqlalchemy import SQLAlchemy
+    from flask_security.models import fsqla_v2 as fsqla
+    from flask_security import Security, SQLAlchemyUserDatastore
+
+    app.config["SQLALCHEMY_DATABASE_URI"] = "sqlite:///:memory:"
+    db = SQLAlchemy(app)
+
+    fsqla.FsModels.set_db_info(db)
+
+    class Role(db.Model, fsqla.FsRoleMixin):
+        pass
+
+    class User(db.Model, fsqla.FsUserMixin):
+        fs_token_uniquifier = Column(String(64), unique=True, nullable=True)
+
+    with app.app_context():
+        db.create_all()
+
+    ds = SQLAlchemyUserDatastore(db, User, Role)
+    app.security = Security(app, datastore=ds)
+
+    with app.app_context():
+        ds.create_user(
+            email="matt@lp.com",
+            password=hash_password("password"),
+        )
+        ds.commit()
+
+        # manually null out fs_token_uniquifier
+        user = ds.find_user(email="matt@lp.com")
+        user.fs_token_uniquifier = None
+        ds.put(user)
+        ds.commit()
+
+        client_nc = app.test_client(use_cookies=False)
+
+        response = json_authenticate(client_nc)
+        token = response.json["response"]["user"]["authentication_token"]
+        verify_token(client_nc, token)
+
+
 def test_token_query(in_app_context):
     # Verify that when authenticating with auth token (and not session)
     # that there is just one DB query to get user.
