@@ -1,4 +1,3 @@
-# -*- coding: utf-8 -*-
 """
     test_misc
     ~~~~~~~~~~~
@@ -12,18 +11,23 @@
 
 from datetime import timedelta
 import hashlib
-import mock
+from unittest import mock
 import re
+import os.path
+import pkg_resources
 import sys
 import time
 
 import pytest
 
-from utils import (
+from tests.test_utils import (
     authenticate,
+    capture_flashes,
+    capture_reset_password_requests,
     check_xlation,
     init_app_with_options,
     json_authenticate,
+    logout,
     populate_data,
 )
 
@@ -32,6 +36,7 @@ from flask_security import Security
 from flask_security.forms import (
     ChangePasswordForm,
     ConfirmRegisterForm,
+    EmailField,
     ForgotPasswordForm,
     LoginForm,
     PasswordField,
@@ -45,60 +50,39 @@ from flask_security.forms import (
     email_validator,
     valid_user_email,
 )
-from flask_security import auth_required
+from flask_security import auth_required, roles_required
 from flask_security.utils import (
-    capture_flashes,
-    capture_reset_password_requests,
     encode_string,
     json_error_response,
+    get_request_attr,
     hash_data,
     send_mail,
-    string_types,
     uia_phone_mapper,
     verify_hash,
 )
 
 
 @pytest.mark.recoverable()
-def test_async_email_task(app, client):
-    app.mail_sent = False
+def test_my_mail_util(app, sqlalchemy_datastore):
+    from flask_security import MailUtil
 
-    @app.security.send_mail_task
-    def send_email(msg):
-        app.mail_sent = True
-
-    client.post("/reset", data=dict(email="matt@lp.com"))
-    assert app.mail_sent is True
-
-
-@pytest.mark.recoverable()
-def test_alt_send_mail(app, sqlalchemy_datastore):
-    """ Verify that can override the send_mail method. """
-    app.mail_sent = False
-
-    def send_email(subject, email, template, **kwargs):
-        app.mail_sent = True
+    class MyMailUtil(MailUtil):
+        def send_mail(
+            self, template, subject, recipient, sender, body, html, user, **kwargs
+        ):
+            assert template == "reset_instructions"
+            assert subject == app.config["SECURITY_EMAIL_SUBJECT_PASSWORD_RESET"]
+            assert recipient == "matt@lp.com"
+            assert user.email == "matt@lp.com"
+            assert sender == "no-reply@localhost"
+            assert isinstance(sender, str)
 
     init_app_with_options(
-        app, sqlalchemy_datastore, **{"security_args": {"send_mail": send_email}}
+        app, sqlalchemy_datastore, **{"security_args": {"mail_util_cls": MyMailUtil}}
     )
+
     client = app.test_client()
-
     client.post("/reset", data=dict(email="matt@lp.com"))
-    assert app.mail_sent is True
-
-
-@pytest.mark.recoverable()
-def test_alt_send_mail_decorator(app, client):
-    """ Verify that can override the send_mail method. """
-    app.mail_sent = False
-
-    @app.security.send_mail
-    def send_email(subject, email, template, **kwargs):
-        app.mail_sent = True
-
-    client.post("/reset", data=dict(email="matt@lp.com"))
-    assert app.mail_sent is True
 
 
 def test_register_blueprint_flag(app, sqlalchemy_datastore):
@@ -113,13 +97,13 @@ def test_register_blueprint_flag(app, sqlalchemy_datastore):
 @pytest.mark.changeable()
 def test_basic_custom_forms(app, sqlalchemy_datastore):
     class MyLoginForm(LoginForm):
-        email = StringField("My Login Email Address Field")
+        email = EmailField("My Login Email Address Field")
 
     class MyRegisterForm(RegisterForm):
-        email = StringField("My Register Email Address Field")
+        email = EmailField("My Register Email Address Field")
 
     class MyForgotPasswordForm(ForgotPasswordForm):
-        email = StringField(
+        email = EmailField(
             "My Forgot Email Address Field",
             validators=[email_required, email_validator, valid_user_email],
         )
@@ -172,10 +156,10 @@ def test_confirmable_custom_form(app, sqlalchemy_datastore):
     app.config["SECURITY_CONFIRMABLE"] = True
 
     class MyRegisterForm(ConfirmRegisterForm):
-        email = StringField("My Register Email Address Field")
+        email = EmailField("My Register Email Address Field")
 
     class MySendConfirmationForm(SendConfirmationForm):
-        email = StringField("My Send Confirmation Email Address Field")
+        email = EmailField("My Send Confirmation Email Address Field")
 
     app.security = Security(
         app,
@@ -197,7 +181,7 @@ def test_passwordless_custom_form(app, sqlalchemy_datastore):
     app.config["SECURITY_PASSWORDLESS"] = True
 
     class MyPasswordlessLoginForm(PasswordlessLoginForm):
-        email = StringField("My Passwordless Email Address Field")
+        email = EmailField("My Passwordless Email Address Field")
 
     app.security = Security(
         app,
@@ -258,17 +242,6 @@ def test_logout_methods_none(app, sqlalchemy_datastore):
     assert response.status_code == 404
 
 
-def test_addition_identity_attributes(app, sqlalchemy_datastore):
-    init_app_with_options(
-        app,
-        sqlalchemy_datastore,
-        **{"SECURITY_USER_IDENTITY_ATTRIBUTES": ("email", "username")}
-    )
-    client = app.test_client()
-    response = authenticate(client, email="matt", follow_redirects=True)
-    assert b"Welcome matt@lp.com" in response.data
-
-
 def test_passwordless_and_two_factor_configuration_mismatch(app, sqlalchemy_datastore):
     with pytest.raises(ValueError):
         init_app_with_options(
@@ -328,7 +301,7 @@ def test_change_hash_type(app, sqlalchemy_datastore):
 
 
 @pytest.mark.settings(hashing_schemes=["hex_md5"], deprecated_hashing_schemes=[])
-@pytest.mark.parametrize("data", [u"hellö", b"hello"])
+@pytest.mark.parametrize("data", ["hellö", b"hello"])
 def test_legacy_hash(in_app_context, data):
     legacy_hash = hashlib.md5(encode_string(data)).hexdigest()
     new_hash = hash_data(data)
@@ -337,23 +310,23 @@ def test_legacy_hash(in_app_context, data):
 
 def test_hash_data(in_app_context):
     data = hash_data(b"hello")
-    assert isinstance(data, string_types)
-    data = hash_data(u"hellö")
-    assert isinstance(data, string_types)
+    assert isinstance(data, str)
+    data = hash_data("hellö")
+    assert isinstance(data, str)
 
 
 def test_verify_hash(in_app_context):
-    data = hash_data(u"hellö")
-    assert verify_hash(data, u"hellö") is True
-    assert verify_hash(data, u"hello") is False
+    data = hash_data("hellö")
+    assert verify_hash(data, "hellö") is True
+    assert verify_hash(data, "hello") is False
 
-    legacy_data = hashlib.md5(encode_string(u"hellö")).hexdigest()
-    assert verify_hash(legacy_data, u"hellö") is True
-    assert verify_hash(legacy_data, u"hello") is False
+    legacy_data = hashlib.md5(encode_string("hellö")).hexdigest()
+    assert verify_hash(legacy_data, "hellö") is True
+    assert verify_hash(legacy_data, "hello") is False
 
 
 @pytest.mark.settings(
-    password_salt=u"öööööööööööööööööööööööööööööööööö", password_hash="bcrypt"
+    password_salt="öööööööööööööööööööööööööööööööööö", password_hash="bcrypt"
 )
 def test_password_unicode_password_salt(client):
     response = authenticate(client)
@@ -437,17 +410,18 @@ def test_form_required_local_message(app, sqlalchemy_datastore):
 
 @pytest.mark.babel(False)
 def test_without_babel(client):
-    response = client.get("/login")
-    assert b"Login" in response.data
+    # This isn't really 'without' babel - it is without initializing babel
+    with pytest.raises(ValueError):
+        client.get("/login")
 
 
 def test_no_email_sender(app):
-    """ Verify that if SECURITY_EMAIL_SENDER is default
-        (which is a local proxy) that send_mail picks up MAIL_DEFAULT_SENDER.
+    """Verify that if SECURITY_EMAIL_SENDER is default
+    (which is a local proxy) that send_mail picks up MAIL_DEFAULT_SENDER.
     """
     app.config["MAIL_DEFAULT_SENDER"] = "test@testme.com"
 
-    class TestUser(object):
+    class TestUser:
         def __init__(self, email):
             self.email = email
 
@@ -455,6 +429,7 @@ def test_no_email_sender(app):
     security.init_app(app)
 
     with app.app_context():
+        app.try_trigger_before_first_request_functions()
         user = TestUser("matt@lp.com")
         with app.mail.record_messages() as outbox:
             send_mail("Test Default Sender", user.email, "welcome", user=user)
@@ -474,6 +449,27 @@ def test_xlation(app, client):
     assert b"Bienvenue matt@lp.com" in response.data
 
 
+def test_myxlation(app, sqlalchemy_datastore, pytestconfig):
+    # Test changing a single MSG and having an additional translation dir
+
+    i18n_dirname = [
+        pkg_resources.resource_filename("flask_security", "translations"),
+        os.path.join(pytestconfig.rootdir, "tests/translations"),
+    ]
+    init_app_with_options(
+        app, sqlalchemy_datastore, **{"SECURITY_I18N_DIRNAME": i18n_dirname}
+    )
+
+    app.config["BABEL_DEFAULT_LOCALE"] = "fr_FR"
+    assert check_xlation(app, "fr_FR"), "You must run python setup.py compile_catalog"
+
+    app.config["SECURITY_MSG_INVALID_PASSWORD"] = ("Password no-worky", "error")
+
+    client = app.test_client()
+    response = client.post("/login", data=dict(email="matt@lp.com", password="forgot"))
+    assert b"Passe - no-worky" in response.data
+
+
 def test_form_labels(app):
     app.config["BABEL_DEFAULT_LOCALE"] = "fr_FR"
     app.security = Security()
@@ -520,6 +516,8 @@ def test_per_request_xlate(app, client):
 
     response = client.get("/login", headers=[("Accept-Language", "fr")])
     assert b'<label for="password">Mot de passe</label>' in response.data
+    # make sure template contents get xlated (not just form).
+    assert b"<h1>Connexion</h1>" in response.data
 
     data = dict(email="matt@lp.com", password="", remember="y")
     response = client.post("/login", data=data, headers=[("Accept-Language", "fr")])
@@ -539,6 +537,7 @@ def test_per_request_xlate(app, client):
     response = client.get("/change", follow_redirects=True)
     assert response.status_code == 200
     assert b"Nouveau mot de passe" in response.data
+    assert b"<h1>Changer de mot de passe</h1>" in response.data
 
     # try JSON
     response = client.post(
@@ -566,8 +565,8 @@ def test_zxcvbn_xlate(app):
 
     with app.test_request_context():
         user = TestUser("jwag@notme.com")
-        result = app.security._password_validator("simple", False, user=user)
-        print(result)
+        pbad, pnorm = app.security._password_util.validate("simple", False, user=user)
+        print(pbad)
 """
 
 
@@ -576,15 +575,13 @@ def test_zxcvbn_xlate(app):
 def test_breached(app):
 
     # partial response from: https://api.pwnedpasswords.com/range/07003
-    pwned_response = "AF5A73CD3CBCFDCD12B0B68CB7930F3E888:2\r\n\
+    pwned_response = b"AF5A73CD3CBCFDCD12B0B68CB7930F3E888:2\r\n\
 AFD8AA47E6FD782ADDC11D89744769F7354:2\r\n\
 B04334E179537C975D0B3C72DA2E5B68E44:15\r\n\
 B118F58C2373FDF97ACF93BD3339684D1EB:2\r\n\
 B1ED5D27429EDF77EFD84F4EA9BDA5013FB:4\r\n\
 B25C03CFBE4CBF19E0F4889711C9A488E5D:2\r\n\
-B3902FD808DCA504AAAD30F3C14BD3ACE7C:10".encode(
-        "utf-8"
-    )
+B3902FD808DCA504AAAD30F3C14BD3ACE7C:10"
 
     app.security = Security()
     app.security.init_app(app)
@@ -593,7 +590,7 @@ B3902FD808DCA504AAAD30F3C14BD3ACE7C:10".encode(
             mock_urlopen.return_value.__enter__.return_value.read.return_value = (
                 pwned_response
             )
-            pbad = app.security._password_validator("flaskflask", False)
+            pbad, pnorm = app.security._password_util.validate("flaskflask", False)
             assert len(pbad) == 1
             assert app.config["SECURITY_MSG_PASSWORD_BREACHED"][0] in pbad[0]
 
@@ -607,15 +604,13 @@ B3902FD808DCA504AAAD30F3C14BD3ACE7C:10".encode(
 def test_breached_cnt(app):
 
     # partial response from: https://api.pwnedpasswords.com/range/07003
-    pwned_response = "AF5A73CD3CBCFDCD12B0B68CB7930F3E888:2\r\n\
+    pwned_response = b"AF5A73CD3CBCFDCD12B0B68CB7930F3E888:2\r\n\
 AFD8AA47E6FD782ADDC11D89744769F7354:2\r\n\
 B04334E179537C975D0B3C72DA2E5B68E44:15\r\n\
 B118F58C2373FDF97ACF93BD3339684D1EB:2\r\n\
 B1ED5D27429EDF77EFD84F4EA9BDA5013FB:4\r\n\
 B25C03CFBE4CBF19E0F4889711C9A488E5D:2\r\n\
-B3902FD808DCA504AAAD30F3C14BD3ACE7C:10".encode(
-        "utf-8"
-    )
+B3902FD808DCA504AAAD30F3C14BD3ACE7C:10"
 
     app.security = Security()
     app.security.init_app(app)
@@ -624,7 +619,7 @@ B3902FD808DCA504AAAD30F3C14BD3ACE7C:10".encode(
             mock_urlopen.return_value.__enter__.return_value.read.return_value = (
                 pwned_response
             )
-            pbad = app.security._password_validator("flaskflask", True)
+            pbad, pnorm = app.security._password_util.validate("flaskflask", True)
             # Still weak password, just not pwned enough. Should fail complexity
             assert len(pbad) == 1
             assert "Repeats like" in pbad[0]
@@ -638,7 +633,7 @@ def test_breached_real(app):
     app.security = Security()
     app.security.init_app(app)
     with app.test_request_context():
-        pbad = app.security._password_validator("flaskflask", True)
+        pbad, pnorm = app.security._password_util.validate("flaskflask", True)
         assert len(pbad) == 1
         assert app.config["SECURITY_MSG_PASSWORD_BREACHED"][0] in pbad[0]
 
@@ -698,7 +693,10 @@ def test_method_view(app, client):
 
 
 def test_phone_util_override(app):
-    class MyPhoneUtil(object):
+    class MyPhoneUtil:
+        def __init__(self, app):
+            pass
+
         def validate_phone_number(self, input_data):
             return "call-me"
 
@@ -709,9 +707,6 @@ def test_phone_util_override(app):
     app.security.init_app(app, phone_util_cls=MyPhoneUtil)
 
     with app.app_context():
-        client = app.test_client()
-        # trigger @before first request
-        client.get("/login")
         assert uia_phone_mapper("55") == "very-canonical"
 
 
@@ -911,3 +906,131 @@ def test_verify_fresh_json(app, client, get_message):
     response = client.get("/fresh", headers=headers)
     assert response.status_code == 200
     assert response.json["title"] == "Fresh Only"
+
+
+@pytest.mark.changeable()
+def test_verify_pwd_json(app, client, get_message):
+    # Make sure verify accepts a normalized and original password.
+    authenticate(client)
+    headers = {"Accept": "application/json", "Content-Type": "application/json"}
+    data = dict(
+        password="password",
+        new_password="new strong password\N{ROMAN NUMERAL ONE}",
+        new_password_confirm="new strong password\N{ROMAN NUMERAL ONE}",
+    )
+    response = client.post(
+        "/change",
+        json=data,
+        headers={"Content-Type": "application/json"},
+    )
+    assert response.status_code == 200
+
+    response = client.post(
+        "/verify",
+        json=dict(password="new strong password\N{ROMAN NUMERAL ONE}"),
+        headers=headers,
+    )
+    assert response.status_code == 200
+
+    response = client.post(
+        "/verify",
+        json=dict(password="new strong password\N{LATIN CAPITAL LETTER I}"),
+        headers=headers,
+    )
+    assert response.status_code == 200
+
+
+@pytest.mark.settings(verify_url="/auth/")
+def test_verify_next(app, client, get_message):
+    authenticate(client)
+    response = client.post(
+        "/auth/?next=http://localhost/mynext",
+        data=dict(password="password"),
+        follow_redirects=False,
+    )
+    assert response.location == "http://localhost/mynext"
+
+    response = client.post(
+        "/auth/?next=http%3A%2F%2F127.0.0.1%3A5000%2Fdashboard%2Fsettings%2F",
+        data=dict(password="password"),
+        follow_redirects=False,
+        base_url="http://127.0.0.1:5000",
+    )
+    assert response.location == "http://127.0.0.1:5000/dashboard/settings/"
+
+
+def test_direct_decorator(app, client, get_message):
+    """ Test/show calling the auth_required decorator directly """
+    headers = {"Accept": "application/json", "Content-Type": "application/json"}
+
+    def myview():
+        return roles_required("author")(domyview)()
+
+    def domyview():
+        return Response(status=200)
+
+    app.add_url_rule("/myview", view_func=myview, methods=["GET"])
+
+    authenticate(client)
+    response = client.get("/myview", headers=headers)
+    assert response.status_code == 403
+
+    logout(client)
+
+    authenticate(client, email="jill@lp.com")
+    response = client.get("/myview", headers=headers)
+    assert response.status_code == 200
+
+
+def test_authn_via(app, client, get_message):
+    """ Test that we get correct fs_authn_via set in request """
+
+    @auth_required(within=30, grace=0)
+    def myview():
+        assert get_request_attr("fs_authn_via") == "session"
+        return Response(status=200)
+
+    app.add_url_rule("/myview", view_func=myview, methods=["GET"])
+    authenticate(client)
+
+    # This should work and not be redirected
+    response = client.get("/myview", follow_redirects=False)
+    assert response.status_code == 200
+
+
+def test_post_security_with_application_root(app, sqlalchemy_datastore):
+    init_app_with_options(app, sqlalchemy_datastore, **{"APPLICATION_ROOT": "/root"})
+    client = app.test_client()
+
+    response = client.post(
+        "/login", data=dict(email="matt@lp.com", password="password")
+    )
+    assert response.status_code == 302
+    assert response.headers["Location"] == "http://localhost/root"
+
+    response = client.get("/logout")
+    assert response.status_code == 302
+    assert response.headers["Location"] == "http://localhost/root"
+
+
+def test_post_security_with_application_root_and_views(app, sqlalchemy_datastore):
+    init_app_with_options(
+        app,
+        sqlalchemy_datastore,
+        **{
+            "APPLICATION_ROOT": "/root",
+            "SECURITY_POST_LOGIN_VIEW": "/post_login",
+            "SECURITY_POST_LOGOUT_VIEW": "/post_logout",
+        }
+    )
+    client = app.test_client()
+
+    response = client.post(
+        "/login", data=dict(email="matt@lp.com", password="password")
+    )
+    assert response.status_code == 302
+    assert response.headers["Location"] == "http://localhost/post_login"
+
+    response = client.get("/logout")
+    assert response.status_code == 302
+    assert response.headers["Location"] == "http://localhost/post_logout"
