@@ -1,4 +1,3 @@
-# -*- coding: utf-8 -*-
 """
     test_registerable
     ~~~~~~~~~~~~~~~~~
@@ -7,24 +6,33 @@
 """
 
 import pytest
+import re
 from flask import Flask
-from utils import authenticate, logout
+import jinja2
+from tests.test_utils import authenticate, check_xlation, logout
 
 from flask_security import Security
 from flask_security.core import UserMixin
-from flask_security.forms import ConfirmRegisterForm, RegisterForm, StringField
+from flask_security.forms import (
+    ConfirmRegisterForm,
+    RegisterForm,
+    StringField,
+    _default_field_labels,
+)
 from flask_security.signals import user_registered
+from flask_security.utils import localize_callback
 
 pytestmark = pytest.mark.registerable()
 
 
 @pytest.mark.settings(post_register_view="/post_register")
-def test_registerable_flag(client, app, get_message):
+def test_registerable_flag(clients, app, get_message):
     recorded = []
 
     # Test the register view
-    response = client.get("/register")
+    response = clients.get("/register")
     assert b"<h1>Register</h1>" in response.data
+    assert re.search(b'<input[^>]*type="email"[^>]*>', response.data)
 
     # Test registering is successful, sends email, and fires signal
     @user_registered.connect_via(app)
@@ -44,37 +52,37 @@ def test_registerable_flag(client, app, get_message):
         next="",
     )
     with app.mail.record_messages() as outbox:
-        response = client.post("/register", data=data, follow_redirects=True)
+        response = clients.post("/register", data=data, follow_redirects=True)
 
     assert len(recorded) == 1
     assert len(outbox) == 1
     assert b"Post Register" in response.data
 
-    logout(client)
+    logout(clients)
 
     # Test user can login after registering
-    response = authenticate(client, email="dude@lp.com", password="battery staple")
+    response = authenticate(clients, email="dude@lp.com", password="battery staple")
     assert response.status_code == 302
 
-    logout(client)
+    logout(clients)
 
     # Test registering with an existing email
     data = dict(
         email="dude@lp.com", password="password", password_confirm="password", next=""
     )
-    response = client.post("/register", data=data, follow_redirects=True)
+    response = clients.post("/register", data=data, follow_redirects=True)
     assert get_message("EMAIL_ALREADY_ASSOCIATED", email="dude@lp.com") in response.data
 
     # Test registering with an existing email but case insensitive
     data = dict(
         email="Dude@lp.com", password="password", password_confirm="password", next=""
     )
-    response = client.post("/register", data=data, follow_redirects=True)
+    response = clients.post("/register", data=data, follow_redirects=True)
     assert get_message("EMAIL_ALREADY_ASSOCIATED", email="Dude@lp.com") in response.data
 
     # Test registering with JSON
     data = dict(email="dude2@lp.com", password="horse battery")
-    response = client.post(
+    response = clients.post(
         "/register", json=data, headers={"Content-Type": "application/json"}
     )
 
@@ -83,17 +91,17 @@ def test_registerable_flag(client, app, get_message):
     assert len(response.json["response"]) == 2
     assert all(k in response.json["response"] for k in ["csrf_token", "user"])
 
-    logout(client)
+    logout(clients)
 
     # Test registering with invalid JSON
     data = dict(email="bogus", password="password")
-    response = client.post(
+    response = clients.post(
         "/register", json=data, headers={"Content-Type": "application/json"}
     )
     assert response.headers["content-type"] == "application/json"
     assert response.json["meta"]["code"] == 400
 
-    logout(client)
+    logout(clients)
 
     # Test ?next param
     data = dict(
@@ -103,10 +111,63 @@ def test_registerable_flag(client, app, get_message):
         next="",
     )
 
-    response = client.post("/register?next=/page1", data=data, follow_redirects=True)
+    response = clients.post("/register?next=/page1", data=data, follow_redirects=True)
     assert b"Page 1" in response.data
 
 
+@pytest.mark.confirmable()
+def test_xlation(app, client, get_message_local):
+    # Test form and email translation
+    app.config["BABEL_DEFAULT_LOCALE"] = "fr_FR"
+    assert check_xlation(app, "fr_FR"), "You must run python setup.py compile_catalog"
+
+    response = client.get("/register", follow_redirects=True)
+    with app.app_context():
+        # Check header
+        assert (
+            f'<h1>{localize_callback("Register")}</h1>'.encode("utf-8") in response.data
+        )
+        submit = localize_callback(_default_field_labels["register"])
+        assert f'value="{submit}"'.encode("utf-8") in response.data
+
+    with app.mail.record_messages() as outbox:
+        response = client.post(
+            "/register",
+            data={
+                "email": "me@fr.com",
+                "password": "new strong password",
+                "password_confirm": "new strong password",
+            },
+            follow_redirects=True,
+        )
+
+    with app.app_context():
+        assert (
+            get_message_local("CONFIRM_REGISTRATION", email="me@fr.com").encode("utf-8")
+            in response.data
+        )
+        assert b"Home Page" in response.data
+        assert len(outbox) == 1
+        assert (
+            localize_callback(app.config["SECURITY_EMAIL_SUBJECT_REGISTER"])
+            in outbox[0].subject
+        )
+        assert (
+            str(
+                jinja2.escape(
+                    localize_callback(
+                        "You can confirm your email through the link below:"
+                    )
+                )
+            )
+            in outbox[0].html
+        )
+        assert (
+            localize_callback("You can confirm your email through the link below:")
+            in outbox[0].body
+        )
+
+
 @pytest.mark.confirmable()
 def test_required_password(client, get_message):
     # when confirm required - should not require confirm_password - but should
@@ -204,7 +265,7 @@ def test_disable_register_emails(client, app):
 @pytest.mark.two_factor()
 @pytest.mark.settings(two_factor_required=True)
 def test_two_factor(app, client):
-    """ If two-factor is enabled, the register shouldn't login, but start the
+    """If two-factor is enabled, the register shouldn't login, but start the
     2-factor setup.
     """
     data = dict(email="dude@lp.com", password="password", password_confirm="password")
@@ -345,3 +406,79 @@ def test_nullable_username(app, sqlalchemy_datastore):
         "/register", json=data, headers={"Content-Type": "application/json"}
     )
     assert response.status_code == 200
+
+
+def test_email_normalization(app, client):
+    # should be able to login either as LP.com or lp.com
+    data = dict(
+        email="\N{BLACK SCISSORS}@LP.com",
+        password="battery staple",
+        password_confirm="battery staple",
+    )
+
+    response = client.post("/register", data=data, follow_redirects=True)
+    assert b"Home Page" in response.data
+    logout(client)
+
+    # Test user can login after registering
+    response = authenticate(
+        client, email="\N{BLACK SCISSORS}@lp.com", password="battery staple"
+    )
+    assert response.status_code == 302
+
+    logout(client)
+    # Test user can login after registering using original non-canonical email
+    response = authenticate(
+        client, email="\N{BLACK SCISSORS}@LP.com", password="battery staple"
+    )
+    assert response.status_code == 302
+
+
+def test_email_normalization_options(app, client, get_message):
+    # verify can set options for email_validator
+    data = dict(
+        email="\N{BLACK SCISSORS}@LP.com",
+        password="battery staple",
+        password_confirm="battery staple",
+    )
+
+    response = client.post("/register", data=data, follow_redirects=True)
+    assert b"Home Page" in response.data
+    logout(client)
+
+    # turn off allowing 'local' part unicode.
+    app.config["SECURITY_EMAIL_VALIDATOR_ARGS"] = {"allow_smtputf8": False}
+    data = dict(
+        email="\N{WHITE SCISSORS}@LP.com",
+        password="battery staple",
+        password_confirm="battery staple",
+    )
+
+    response = client.post("/register", data=data, follow_redirects=True)
+    assert response.status_code == 200
+    assert get_message("INVALID_EMAIL_ADDRESS") in response.data
+
+
+def test_form_error(app, client, get_message):
+    # A few form validations use ValidatorMixin which provides a lazy string
+    # Since CLI doesn't render_template it was seeing those lazy strings.
+    # This test basically just illustrates all this.
+    from babel.support import LazyProxy
+
+    with app.test_request_context("/register"):
+        # this is usually done @before_first_request
+        app.jinja_env.globals["_fsdomain"] = app.security.i18n_domain.gettext
+        rform = ConfirmRegisterForm()
+
+        rform.validate()
+        assert isinstance(rform.errors["email"][0], LazyProxy)
+        assert str(rform.errors["email"][0]).encode("utf-8") == get_message(
+            "EMAIL_NOT_PROVIDED"
+        )
+
+        # make sure rendered template has converted LocalProxy strings.
+        rendered = app.security.render_template(
+            app.config["SECURITY_REGISTER_USER_TEMPLATE"],
+            register_user_form=rform,
+        )
+        assert get_message("EMAIL_NOT_PROVIDED") in rendered.encode("utf-8")
